'''
source code for likelihoods etc for manual models
'''

from jax import numpy as jnp, lax, random, custom_jvp

from numpyro.distributions import constraints, Normal, HalfNormal
from numpyro.distributions.transforms import AffineTransform
from numpyro.distributions.distribution import Distribution, TransformedDistribution
from numpyro.distributions.util import (
    promote_shapes,
    validate_sample
)


def reshape_times_and_kappa(ti, kap):
    ## fix minimal number of dimensions
    if ti.ndim == 0:
        ti = ti.reshape((-1,))
    if kap.size == 1:
        kap = jnp.ones_like(ti) * kap
    return ti, kap

# define power-generalized weibull functions with analytic gradients (for numerical stability)
@custom_jvp
def pgwh0(ti,kap):
    '''
    baseline hazard
    
    :param ti: vector of times
    :param kap: kappa, single number or vector of length ti
    '''
    out = (1+ti/(kap+1))**(kap-1)
    ## fix infinite cases
    out = jnp.where(jnp.isinf(kap), jnp.exp(ti), out)
    
    return out

def pgwh0d(ti,kap):
    'time derivative of hazard function'
    out = pgwh0(ti,kap)*pgwm0d(ti,kap)
    return out

def pgwh0k(ti,kap):
    'kappa derivative of hazard function'
    out = pgwh0(ti,kap)*pgwm0k(ti,kap)
    return out

@custom_jvp
def pgwm0(ti,kap):
    'log hazard function'
    # ti, kap = reshape_times_and_kappa(ti, kap)
    out = (kap - 1)*jnp.log(1 + (ti / (kap + 1)))
    out = jnp.where(jnp.isinf(kap), ti, out)
    return out

def pgwm0d(ti,kap):
    'z-derivative of log-hazard function'
    out = (kap-1) / (kap + 1 + ti)
    out = jnp.where(jnp.isinf(kap), 1, out)
    return out

def pgwm0k(ti,kap):
    'kappa-derivative of log hazard function'
    out = pgwm0(ti,kap) / (kap - 1) - ti * pgwm0d(ti,kap) / (kap + 1)
    out = jnp.where(kap==0, 0, out)
    out = jnp.where(jnp.isinf(kap), 0, out)
    return out

@custom_jvp
def pgwH0(ti,kap):
    '''
    baseline cumulative hazard
    
    :param ti: vector of times
    :param kap: kappa, single number or vector of length ti
    '''
    
    k0 = kap == 0
    ## replace zeros with small values for numeric stability (fix this later)
    kap = jnp.where(k0, 1e-6, kap)
    out = ((kap+1)/kap)*((1+ti/(kap+1))**kap-1)
    ## replace zeros sites with correct calculation
    out = jnp.where(k0, jnp.log(1+ti), out)
    ## fix infinite cases
    out = jnp.where(jnp.isinf(kap), jnp.exp(ti) - 1, out)

    return out

def pgwH0k(ti,kap):
    'k-derivative of cumulative hazard function'
    # ti, kap = reshape_times_and_kappa(ti,kap)
    H0i = pgwH0(ti,kap)
    h0i = pgwh0(ti,kap)
    m0i = pgwm0(ti,kap)
    out = ((kap*H0i+kap+1)*m0i)/(kap*(kap-1)) - H0i/(kap*(kap+1)) - ti*h0i/(kap+1)
    out = jnp.where(kap==0, 0, out)
    out = jnp.where(jnp.isinf(kap), 0, out)
    return out

def pgwH0d(ti,kap):
    'ti-derivative of cumulative hazard function'
    # ti, kap = reshape_times_and_kappa(ti,kap)
    out = (ti / (kap + 1) + 1)**(kap - 1)
    out = jnp.where(jnp.isinf(kap), 0, out)
    return out

# now define the custom analysical derivatives
@pgwH0.defjvp
def pgwH0_jvp(primals, tangents):
    ti, kap = primals
    ti_dot, kap_dot = tangents
    primal_out  = pgwH0(ti, kap)
    tangent_out = pgwH0d(ti, kap) * ti_dot + pgwH0k(ti, kap) * kap_dot
    return primal_out, tangent_out

# analytical derivative of h0
@pgwh0.defjvp
def pgwh0_jvp(primals, tangents):
    ti, kap = primals
    ti_dot, kap_dot = tangents
    primal_out  = pgwh0(ti, kap)
    tangent_out = pgwh0d(ti, kap) * ti_dot + pgwh0k(ti, kap) * kap_dot
    return primal_out, tangent_out

# analytical derivative of m0
@pgwm0.defjvp
def pgwm0_jvp(primals, tangents):
    ti, kap = primals
    ti_dot, kap_dot = tangents
    primal_out  = pgwm0(ti, kap)
    tangent_out = pgwm0d(ti, kap) * ti_dot + pgwm0k(ti, kap) * kap_dot
    return primal_out, tangent_out

class PowerGeneralizedWeibull(Distribution):
    '''
    PowerGeneralizedWeibull distributions
    see https://arxiv.org/abs/1901.03212
    '''
    arg_constraints = {'phi': constraints.positive, 
                       'lam': constraints.positive, 
                       'gam': constraints.positive,
                       'kap': constraints.greater_than(-1.0)}
    support = constraints.real
    reparametrized_params = ['phi', 'lam', 'gam', 'kap']

    def __init__(self, phi=1., lam=1., gam=1., kap=-0.1, validate_args=None):
        '''
        note that all parameters are expected to be exponentiated if they stem from a (linear) regression model,
        moreover, kappa should be substracted by 1 to allow for negative values of kappa    

        :param phi: part depending on tau-regression (Accelerated Failure Time)
        :param lam: part depending on beta-regression (Proportional Hazards)
        :param gam: part depending on alpha-regression (shape parameter (allows for non-proportional hazards))
        :param kap: part depending on nu-regression    (shape parameter for baseline hazard)
        '''
        self.phi, self.lam, self.gam, self.kap = promote_shapes(phi, lam, gam, kap)
        if not isinstance(self.phi, jnp.ndarray):
            self.phi = jnp.array(self.phi, dtype=jnp.float32)
        if not isinstance(self.lam, jnp.ndarray):
            self.lam = jnp.array(self.lam, dtype=jnp.float32)
        if not isinstance(self.gam, jnp.ndarray):
            self.gam = jnp.array(self.gam, dtype=jnp.float32)
        if not isinstance(self.kap, jnp.ndarray):
            self.kap = jnp.array(self.kap, dtype=jnp.float32)
        
        batch_shape = lax.broadcast_shapes(jnp.shape(phi), jnp.shape(lam), jnp.shape(gam), jnp.shape(kap))
        super(PowerGeneralizedWeibull, self).__init__(batch_shape=batch_shape, validate_args=validate_args)

    def sample(self, key, sample_shape=()):
        u = random.uniform(key, shape=sample_shape + self.batch_shape)
        return (1/self.phi)*((self._pgwH0(-jnp.log(u)/self.lam, 1/self.kap))**(1/self.gam))

    @validate_sample
    def log_prob(self, value):
        # :param value: the observed values; currently to indicate censoring, 
        # values < 0 are considered censored (so values = [0.3, -0.2] <==> ({0.3, 0.2}, {1, 0}) in standard survival analysis notation)

        # find out which observations are censored, by convention: values < 0 are considered censored
        if not isinstance(value, jnp.ndarray):
            value = jnp.array(value)
        notcensored = value >= 0
        times       = jnp.abs(value)

        # calculate args
        zi  = (self.phi*times)**self.gam
        zi, kap = reshape_times_and_kappa(zi, self.kap)
        h0i = self._pgwh0(zi, kap)
        H0i = self._pgwH0(zi, kap)
        
        return notcensored*jnp.log((self.lam*self.gam*zi/times)*h0i) - self.lam*H0i

    @property
    def mean(self):
        return jnp.full(self.batch_shape, jnp.nan)

    @property
    def variance(self):
        return jnp.full(self.batch_shape, jnp.nan)
    
    def _pgwH0(self,ti,kap=None):
        '''
        cumulative hazard function
        '''
        kap = self.kap if kap is None else kap
        return pgwH0(ti,kap)

    def _pgwh0(self,ti,kap=None):
        '''
        hazard function
        '''
        kap = self.kap if kap is None else kap
        return pgwh0(ti,kap)

    def survival(self, times):
        '''
        cumulative survival function
        '''
        if not isinstance(times, jnp.DeviceArray):
            times = jnp.array(times)
        return jnp.exp(-self.lam*self._pgwH0((self.phi*times)**self.gam))

    def hazard(self, times):
        '''
        hazard function
        '''
        if not isinstance(times, jnp.DeviceArray):
            times = jnp.array(times)
        return self._pgwh0((self.phi*times)**self.gam)*self.lam*self.gam*self.phi*((self.phi*times)**(self.gam-1))

def PowerGeneralizedWeibullLog(tau=0.0, beta=0.0, alpha=0.0, nu=0.0, validate_args=None):
    '''
    the PowerGeneralizedWeibull distribution, parameterized with log parameters with unrestricted scale
    '''
    phi = jnp.exp(tau)
    lam = jnp.exp(beta)
    gam = jnp.exp(alpha)
    kap = jnp.exp(nu) - 1.0

    return PowerGeneralizedWeibull(phi, lam, gam, kap, validate_args=validate_args)


class NegativeHalfNormal(TransformedDistribution):
    arg_constraints = {'scale': constraints.positive}
    support = constraints.less_than_eq(0.0)
    reparametrized_params = ["scale"]

    def __init__(self, scale=1., validate_args=None):
        base_dist = HalfNormal(scale)
        self.scale = base_dist.scale
        super(NegativeHalfNormal, self).__init__(
            base_dist, AffineTransform(0, -1), validate_args=validate_args
        )

    @property
    def mean(self):
        return -self.base_dist.mean
    
    @property
    def variance(self):
        return self.base_dist.variance


class DummyNegativeHalfNormal(Distribution):
    '''
    this is just a HalfNormal distribution but then renamed
    '''
    reparametrized_params = ['scale']
    support = constraints.positive
    arg_constraints = {'scale': constraints.positive}

    def __init__(self, scale=1., validate_args=None):
        self._normal = Normal(0., scale)
        self.scale = scale
        super(DummyNegativeHalfNormal, self).__init__(batch_shape=jnp.shape(scale), validate_args=validate_args)

    def sample(self, key, sample_shape=()):
        return jnp.abs(self._normal.sample(key, sample_shape))

    @validate_sample
    def log_prob(self, value):
        return self._normal.log_prob(value) + jnp.log(2)

    @property
    def mean(self):
        return jnp.sqrt(2 / jnp.pi) * self.scale

    @property
    def variance(self):
        return (1 - 2 / jnp.pi) * self.scale ** 2

        



"""
testing for model checks
"""

import numpy as np
import jax
import numpyro
numpyro.set_platform("cpu")
numpyro.set_host_device_count(16)
import pytest

import pandas as pd
from jax import numpy as jnp, random

from causalprotect.inference import PROTECTInference
from causalprotect.models import PROTECTModel
from causalprotect.utils import load_yaml, summary_likelihoods_to_df

rng_key = random.PRNGKey(0)
NUM_OBS = 30
NUM_FOLDS = 2
NUM_GLOBAL_SAMPLES = 10
BASELINE_MCMC_KWARGS = {"num_warmup": 50, "num_samples": 50, "num_chains": 2, "chain_method": "sequential"}
INFERENCE_MCMC_KWARGS = {"num_warmup": 50, "num_samples": 50, "num_chains": 2, "chain_method": "sequential"}
POSTPRED_MCMC_KWARGS = {"num_warmup": 50, "num_samples": 50}

@pytest.fixture(scope="module")
def load_data():
    # read the simulated data
    data_pd = pd.read_csv("tests/testdata.csv")
    data_pd_head = data_pd.head(NUM_OBS)
    data = {k: np.array(v) for k, v in data_pd_head.items()}
    return data


@pytest.fixture(scope="module")
def load_metadata():
    return load_yaml("tests/metadata.yaml")


@pytest.fixture(scope="module")
def load_protect_model(load_data, load_metadata):
    from model import tutorial_model

    return PROTECTModel(
        tutorial_model, load_metadata, prior_spec="tests/priors.csv", data=load_data
    )

@pytest.fixture(scope="module")
def protector(load_data, load_protect_model):
    return PROTECTInference(load_protect_model, load_data)

@pytest.fixture(scope="module")
def test_model_checks_baseline_only(protector):

    result, in_train_mat = protector.model_checks(
        rng_key, num_folds = NUM_FOLDS, 
        verbose=True,
        do_baseline=True,
        do_postpred=False,
        num_global_samples=10,
        baseline_mcmc_kwargs=BASELINE_MCMC_KWARGS,
    )

    exected_pp_modes = {
        "baseline",
    }

    assert set(result.keys()) == exected_pp_modes

    return result, in_train_mat


@pytest.fixture(scope="module")
def test_model_checks_no_baseline(protector):

    result, in_train_mat = protector.model_checks(
        rng_key, num_folds = NUM_FOLDS, 
        verbose=True,
        do_baseline=False,
        do_postpred=True,
        num_global_samples=NUM_GLOBAL_SAMPLES,
        inference_mcmc_kwargs=INFERENCE_MCMC_KWARGS,
        postpred_mcmc_kwargs=POSTPRED_MCMC_KWARGS,
    )

    exected_pp_modes = {
        "joint",
        "no_y",
        "no_tx",
        "no_txy",
        "no_proxy1",
        "no_proxy1_no_y",
        "no_proxy1_no_tx",
        "no_proxy2",
        "no_proxy2_no_y",
        "no_proxy2_no_tx",
    }


    assert set(result.keys()) == exected_pp_modes

    return result, in_train_mat


@pytest.fixture(scope="module")
def test_model_checks_all(protector):

    result, in_train_mat = protector.model_checks(
        rng_key, num_folds = NUM_FOLDS, 
        verbose=True,
        do_baseline=True,
        # do_postpred=True,
        num_global_samples=NUM_GLOBAL_SAMPLES,
        baseline_mcmc_kwargs=BASELINE_MCMC_KWARGS,
        inference_mcmc_kwargs=INFERENCE_MCMC_KWARGS,
        postpred_mcmc_kwargs=POSTPRED_MCMC_KWARGS,
    )

    exected_pp_modes = {
        "baseline",
        "joint",
        "no_y",
        "no_tx",
        "no_txy",
        "no_proxy1",
        "no_proxy1_no_y",
        "no_proxy1_no_tx",
        "no_proxy2",
        "no_proxy2_no_y",
        "no_proxy2_no_tx",
    }

    assert set(result.keys()) == exected_pp_modes

    return result, in_train_mat

def test_model_check_evaluation(
        load_metadata,
        protector,
        test_model_checks_all,
):
    model_check_result, _ = test_model_checks_all
    model_check_evaluation = protector.protect_model.model_check_evaluation(model_check_result)
    f_proxies = load_metadata['f_proxies']
    obs_vars = f_proxies + ['y', 'tx']
    set1 = model_check_evaluation['set1']
    for obs_var in obs_vars:
        assert obs_var in set1, f"{obs_var} not in set1"
        assert 'passed' in set1[obs_var]
    
    set2 = model_check_evaluation['set2']
    for proxy in f_proxies:
        assert proxy in set2, f"{proxy} not in set2"
        proxy_res = set2[proxy]
        assert 'no_proxy' in proxy_res, f"'no_{proxy}' not in {proxy_res}"
        assert 'no_proxy_no_tx'in proxy_res
        assert 'no_proxy_no_y'in proxy_res
        assert 'passed' in proxy_res

    assert 'all_passed' in model_check_evaluation, "'all_passed' not in model_check_evaluation"

    # make the dataframe
    from causalprotect.utils import model_check_evaluation_to_df
    df = model_check_evaluation_to_df(model_check_evaluation)
    # check that the dataframe has the expected columns
    assert 'passed' in df.columns, "'passed' not in df columns"
    # check corret number of rows
    assert len(df) == len(obs_vars) + len(f_proxies), f"Expected {len(obs_vars) + len(f_proxies)} rows, got {len(df)}"



def test_pp_seed_consistency(
    test_model_checks_baseline_only,
    test_model_checks_no_baseline,
    test_model_checks_all,
):

    baseline_only_ll, baseline_only_intrain = test_model_checks_baseline_only
    no_baseline_ll, no_baseline_intrain = test_model_checks_no_baseline
    all_ll, all_intrain = test_model_checks_all

    # check that the in_train matrices are all the same
    assert np.allclose(baseline_only_intrain, all_intrain)
    assert np.allclose(no_baseline_intrain, all_intrain)

    # check that the likelihoods are all the same
    # check that the postpred results are consistent when using a single seed
    assert all_ll['baseline'][('test', 'joint')]['max'][0] == pytest.approx(
        baseline_only_ll['baseline'][('test', 'joint')]['max'][0],
        abs = 0.0001
    )
    assert all_ll['no_y'][('test', 'joint')]['max'][0] == pytest.approx(
        no_baseline_ll['no_y'][('test', 'joint')]['max'][0],
        abs = 0.0001
    )


def test_model_checks_mcmc(protector):

    result, in_train_mat = protector.model_checks(
        rng_key, num_folds = NUM_FOLDS, 
        verbose=True,
        do_baseline=False,
        do_postpred=['no_y'],
        pp_inference='mcmc',
        num_global_samples=NUM_GLOBAL_SAMPLES,
        inference_mcmc_kwargs=INFERENCE_MCMC_KWARGS,
        postpred_mcmc_kwargs=POSTPRED_MCMC_KWARGS,
    )

    exected_pp_modes = {
        "no_y",
    }


    assert set(result.keys()) == exected_pp_modes




"""
test the post pred methods
"""

import pytest

import numpy as np
import numpyro

import pandas as pd
from jax import random

from causalprotect.models import PROTECTModel, tutorial_model
from causalprotect.inference import PROTECTInference
from causalprotect.utils import load_yaml

RNG_SEED = 123
NUM_CHAINS = 2
NUM_LOCAL_DRAWS = 5

numpyro.set_host_device_count(NUM_CHAINS)

@pytest.fixture(scope="module")
def rng_key():
    rng_key = random.PRNGKey(RNG_SEED)
    return rng_key


# read the simulated data
@pytest.fixture(scope="module")
def data():
    data_pd = pd.read_csv("tests/testdata.csv")
    data = {k: np.array(v) for k, v in data_pd.items()}
    return data


@pytest.fixture(scope="module")
def model(data):
    metadata = load_yaml("tests/metadata.yaml")

    model = PROTECTModel(
        tutorial_model, metadata, prior_spec="tests/priors.csv", data=data
    )
    model.check_data(data)
    return model


@pytest.fixture(scope="module")
def protector(data, model):
    protector = PROTECTInference(model, data)
    return protector

def test_postpred(protector, rng_key):
    nobs = protector.default_control["N"]
    rng_key, sub_key = random.split(rng_key)
    # first run inference
    global_samples = protector.run_inference(
        rng_key=sub_key,
        mcmc_kwargs = dict(
            num_samples=50,
            num_warmup=10,
            num_chains=NUM_CHAINS,
            chain_method="parallel",
            progress_bar=False,
        )
    )
    
    # test that the shape of the samples is correct for no_tx ppmode
    rng_key, sub_key = random.split(rng_key)
    pp_samples_no_tx = protector.run_postpred_mcmc(
        sub_key,
        global_samples,
        pp_mode="no_tx",
        num_local_draws=NUM_LOCAL_DRAWS,
        group_by_chain=False,
        mcmc_kwargs={
            "num_samples": 50,
            "num_warmup": 10,
        },
    )

    # check shapes of samples
    assert pp_samples_no_tx['alpha0'].shape == (NUM_LOCAL_DRAWS, )
    assert pp_samples_no_tx['lp'].shape == (NUM_LOCAL_DRAWS, 2, nobs)
    
    ## test per-chain
    global_samples_by_chain = protector.mcmc.get_samples(group_by_chain=True)

    rng_key, sub_key = random.split(rng_key)
    pp_samples_by_chain = protector.run_postpred_mcmc(
        sub_key,
        global_samples_by_chain,
        num_local_draws=NUM_LOCAL_DRAWS,
        group_by_chain=True,
        mcmc_kwargs={
            "num_samples": 50,
            "num_warmup": 10,
            "num_chains": NUM_CHAINS,
            "chain_method": "parallel",
            "progress_bar": False,
        },
    )

    # check shapes of samples
    assert pp_samples_by_chain['alpha0'].shape == (NUM_CHAINS, NUM_LOCAL_DRAWS // NUM_CHAINS, )
    assert pp_samples_by_chain['Feps'].shape == (NUM_CHAINS, NUM_LOCAL_DRAWS // NUM_CHAINS, nobs)

    ## test not per-chain
    rng_key, sub_key = random.split(rng_key)
    pp_samples = protector.run_postpred_mcmc(
        sub_key,
        global_samples,
        num_local_draws=NUM_LOCAL_DRAWS,
        group_by_chain=False,
        mcmc_kwargs={
            "num_samples": 50,
            "num_warmup": 10,
        },
    )

    # check shapes of samples
    assert pp_samples['alpha0'].shape == (NUM_LOCAL_DRAWS, )
    assert pp_samples['Feps'].shape == (NUM_LOCAL_DRAWS, nobs)




import math
import jax
import pytest
import numpy as np
import jax.numpy as jnp

from causalprotect.utils import (
    harrell_c,
    harrell_c_streaming,
    roc_auc,
)


@pytest.mark.parametrize("y, s, expected_auc", [
    # Perfect separation
    (jnp.array([0,0,1,1]), jnp.array([0.1,0.2,0.8,0.9]), 1.0),
    # Complete misranking
    (jnp.array([0,0,1,1]), jnp.array([0.9,0.8,0.2,0.1]), 0.0),
    # Classic small example
    (jnp.array([0,0,1,1,0,1]), jnp.array([0.1,0.4,0.35,0.8,0.5,0.6]), 0.7777777777777778),
    # Ties average to 0.5
    (jnp.array([0,1,0,1]), jnp.array([0.5,0.5,0.2,0.2]), 0.5),
])
def test_auc_unweighted_cases(y, s, expected_auc):
    auc = float(roc_auc(y, s))
    assert auc == pytest.approx(expected_auc, rel=1e-6)

def test_auc_weighted_integer_weights():
    y = jnp.array([0,1,0,1])
    s = jnp.array([0.5,0.5,0.5,0.1])
    w = jnp.array([1,3,1,1], dtype=jnp.float32)
    auc = float(roc_auc(y, s, sample_weight=w))
    assert auc == pytest.approx(0.375, rel=1e-6)

def test_auc_weighted_mixed():
    y = jnp.array([0,0,1,1,0,1])
    s = jnp.array([0.1,0.4,0.35,0.8,0.5,0.6])
    w = jnp.array([1,2,1,1,1,3], dtype=jnp.float32)
    auc = float(roc_auc(y, s, sample_weight=w))
    assert auc == pytest.approx(0.85, rel=1e-6)

def test_degenerate_all_negative_returns_nan():
    y = jnp.array([0,0,0,0])
    s = jnp.array([0.1,0.2,0.3,0.4])
    auc = float(roc_auc(y, s))
    assert np.isnan(auc)

def test_degenerate_all_positive_returns_nan():
    y = jnp.array([1,1,1,1])
    s = jnp.array([0.1,0.2,0.3,0.4])
    auc = float(roc_auc(y, s))
    assert np.isnan(auc)

def test_batch_with_vmap():
    Ys = jnp.stack([jnp.array([0,0,1,1]), jnp.array([0,0,1,1])])
    Ss = jnp.stack([jnp.array([0.1,0.2,0.8,0.9]), jnp.array([0.9,0.8,0.2,0.1])])

    aucs = jax.vmap(roc_auc)(Ys, Ss)
    assert aucs[0] == pytest.approx(1.0)
    assert aucs[1] == pytest.approx(0.0)


## tests for survival concordance index

def test_perfect_concordance():
    T = jnp.array([1., 2., 3., 4.])
    S = jnp.array([4., 3., 2., 1.])  # higher risk earlier -> perfect
    E = jnp.array([1, 1, 1, 1], dtype=bool)
    c = float(harrell_c(T, S, E))
    assert c == pytest.approx(1.0, abs=1e-6)

def test_anti_concordance():
    T = jnp.array([1., 2., 3., 4.])
    S = jnp.array([1., 2., 3., 4.])  # lower risk earlier -> worst
    E = jnp.array([1, 1, 1, 1], dtype=bool)
    c = float(harrell_c(T, S, E))
    assert c == pytest.approx(0.0, abs=1e-6)

def test_ties_count_as_half():
    # Pairs: (0,1)=tie -> 0.5; (0,2)=1>0 ->1; (1,2)=1>0 ->1  => (0.5+1+1)/3 = 0.833333...
    T = jnp.array([1., 2., 3.])
    S = jnp.array([1., 1., 0.])
    E = jnp.array([1, 1, 1], dtype=bool)
    c = float(harrell_c(T, S, E))
    assert c == pytest.approx(2.5/3.0, abs=1e-6)

def test_censoring_rules_out_early_censored_as_defining_event():
    # i=0 is censored at time 2 -> cannot define comparable pairs
    # Only i=1 (event at 3) compares to j=2 (time 4): S[1]=0.2 > S[2]=0.1 -> concordant
    T = jnp.array([2., 3., 4.])
    E = jnp.array([0, 1, 1], dtype=bool)
    S = jnp.array([0.9, 0.2, 0.1])
    c = float(harrell_c(T, S, E))
    assert c == pytest.approx(1.0, abs=1e-6)

def test_no_comparable_pairs_returns_nan():
    # All censored -> no comparable pairs
    T = jnp.array([1., 1.])
    S = jnp.array([0.5, 0.7])
    E = jnp.array([0, 0], dtype=bool)
    c = float(harrell_c(T, S, E))
    assert math.isnan(c)

def test_streaming_matches_vectorized():
    # Deterministic small case
    T = jnp.array([5., 6., 6., 2., 4., 3.])
    E = jnp.array([1, 0, 1, 1, 0, 1], dtype=bool)
    S = jnp.array([0.2, 0.1, 0.9, 0.8, 0.6, 0.4])
    c1 = float(harrell_c(T, S, E))
    c2 = float(harrell_c_streaming(T, S, E))
    if math.isnan(c1) and math.isnan(c2):
        assert True
    else:
        assert c1 == pytest.approx(c2, abs=1e-6)

def test_shape_mismatch_raises():
    T = jnp.array([1., 2., 3.])
    S = jnp.array([0.1, 0.2])  # wrong length
    E = jnp.array([1, 1, 1], dtype=bool)
    with pytest.raises(ValueError):
        _ = harrell_c(T, S, E)

def test_survival_like_outputs_need_sign_flip():
    # Model outputs "higher=better" (e.g., survival prob). Flipping sign fixes it.
    T = jnp.array([1., 2., 3., 4.])
    survival_like = jnp.array([0.1, 0.2, 0.3, 0.4])  # higher later -> better later
    E = jnp.array([1, 1, 1, 1], dtype=bool)

    c_bad = float(harrell_c(T, survival_like, E))         # should be ~0.0
    c_good = float(harrell_c(T, -survival_like, E))       # should be ~1.0

    assert c_bad == pytest.approx(0.0, abs=1e-6)
    assert c_good == pytest.approx(1.0, abs=1e-6)



# test marginalization of hazard ratio in power generalized weibull model
# compare with tests/rcode/marginal_hr.R

from causalprotect.inference import _marginalize_hazard_ratio_pgw
import pytest


import jax
from jax import numpy as jnp
from jax import random

RNG_SEED = 123
N = int(1e5)
B_TX_Y = 1.0

@pytest.fixture(scope="module")
def sample():
    rng_key = random.PRNGKey(RNG_SEED)

    lp_notx = random.normal(rng_key, (N,))
    lp_dotx = lp_notx + B_TX_Y

    sample = {
        "beta0": -2.3,
        "alpha0": -0.01,
        "nu0": 0.7,
        "lp_notx": lp_notx,
        "lp_dotx": lp_dotx,
        "b_tx_y": B_TX_Y
    }
    return sample

def test_marginalize_hazard_ratio_pgw(sample):
    rng_key = random.PRNGKey(RNG_SEED)
    
    # Test with a single sample
    result, t0, t1 = _marginalize_hazard_ratio_pgw(rng_key, sample, maxtime=100)
    assert result.converged
    assert result.position.shape == (4,)
    assert t0.shape == (N,)
    assert t1.shape == (N,)
    assert result.position[1] == pytest.approx(0.642, abs=2e-2)




"""
testing for main module: protector
"""

import arviz as az
import numpy as np
import jax
import numpyro
numpyro.set_platform("cpu")
numpyro.set_host_device_count(4)
import pytest

import pandas as pd
from jax import numpy as jnp, random

k = random.PRNGKey(0)
from causalprotect.inference import PROTECTInference
from causalprotect.models import PROTECTModel
from causalprotect.utils import load_yaml, summary_likelihoods_to_df

NUM_OBS = 30
NUM_FOLDS = 2
NUM_CHAINS = 2
NUM_SAMPLES = 50
NUM_LOCAL_DRAWS = 20

def test_slice_posterior_for_pp():
    from causalprotect.inference import _slice_posterior_for_pp

    num_chains = 2
    num_samples = 20
    num_samples_out = 10

    # test with group_by_chain=False
    samples = {"beta0": jnp.arange(num_chains * num_samples)}
    global_sample_shape = (num_chains * num_samples,)
    sliced = _slice_posterior_for_pp(samples, global_sample_shape, num_samples_out=num_samples_out, group_by_chain=False)
    assert sliced["beta0"].shape == (num_samples_out,)

    # test with group_by_chain=True
    samples = {"beta0": jnp.arange(num_chains * num_samples).reshape((num_chains, num_samples))}
    global_sample_shape = (num_chains, num_samples)
    sliced = _slice_posterior_for_pp(samples, global_sample_shape, num_samples_out=num_samples_out, group_by_chain=True)
    assert sliced["beta0"].shape == (num_chains, num_samples_out // num_chains)

    # test that it throws an error if num_samples_out is bigger than num_chains * num_samples
    with pytest.raises(ValueError):
        _slice_posterior_for_pp(samples, global_sample_shape, num_samples_out=num_chains * num_samples + 1, group_by_chain=False)
    # test that it throws an error if num_samples_out is not divisible by num_chains
    with pytest.raises(ValueError):
        _slice_posterior_for_pp(samples, global_sample_shape, num_samples_out=num_chains * num_samples + 1, group_by_chain=True)



@pytest.fixture(scope="module")
def load_data():
    # read the simulated data
    data_pd = pd.read_csv("tests/testdata.csv")
    data_pd_head = data_pd.head(NUM_OBS)
    data = {k: np.array(v) for k, v in data_pd_head.items()}
    return data


@pytest.fixture(scope="module")
def load_metadata():
    return load_yaml("tests/metadata.yaml")


@pytest.fixture(scope="module")
def load_protect_model(load_data, load_metadata):
    from model import tutorial_model

    return PROTECTModel(
        tutorial_model, load_metadata, prior_spec="tests/priors.csv", data=load_data
    )


def test_data(load_data, load_protect_model):
    assert load_protect_model.check_data(load_data)



@pytest.fixture(scope="module")
def protector(load_data, load_protect_model):
    return PROTECTInference(load_protect_model, load_data)


class TestProtectInference:
    @pytest.fixture(scope="class")
    def inference_samples(self, protector):
        rng_key = random.PRNGKey(6)
        samples = protector.run_inference(
            rng_key,
            mcmc_kwargs={"num_warmup": 50, "num_samples": NUM_SAMPLES, "num_chains": NUM_CHAINS},
        )
        return samples

    def test_inference(self, inference_samples):
        expected_prm_names = [
            "Feps",
            "Fhat",
            "alpha0",
            "b_F_proxy1",
            "b_F_proxy2",
            "b_F_tx",
            "b_F_y",
            "b_Ftx_y",
            "b_age_F",
            "b_tx_y",
            "b_tx_y_marginal",
            "beta0",
            "eta_tx",
            "lp",
            "lp_dotx",
            "lp_notx",
            "mu_proxy1",
            "mu_proxy2",
            "mu_tx",
            "nu0",
            "F_mu",
            "F_sd",
        ]
        assert set(inference_samples.keys()) == set(expected_prm_names)
        assert inference_samples["b_tx_y"].shape == (NUM_CHAINS * NUM_SAMPLES, )

    def test_get_samples(self, protector, inference_samples):
        samples = protector.get_samples()
        assert samples is not None

    @pytest.fixture(scope="class")
    def postpred_noy(self, protector, inference_samples):
        postpred_samples = protector.run_postpred_mcmc(random.PRNGKey(7), inference_samples, pp_mode="no_y", num_local_draws=NUM_LOCAL_DRAWS)
        assert postpred_samples is not None
        assert set(postpred_samples.keys()) == set(inference_samples.keys())
        assert postpred_samples["b_tx_y"].shape == (NUM_LOCAL_DRAWS,)
        return postpred_samples


    def test_c_index(self, protector, postpred_noy):

        c_index = protector.get_c_index()

        assert c_index is not None
        assert c_index.shape == (NUM_LOCAL_DRAWS, )
        assert c_index.mean() > 0
        assert c_index.mean() < 1
